﻿WoWUnit = {}

function WoWUnit:GetMocks()
	self.mocks = self.mocks or {};
	return self.mocks;
end

function WoWUnit:GetTestSuites()
	self.suites = self.suites or {};
	return self.suites;
end

local
function _copyValue(object)
    local lookup_table = {}
    local function _copy(object)
        if type(object) ~= "table" then
            return object
        elseif lookup_table[object] then
            return lookup_table[object]
        end
        local new_table = {}
        lookup_table[object] = new_table
        for index, value in pairs(object) do
            new_table[_copy(index)] = _copy(value)
        end
        return setmetatable(new_table, _copy(getmetatable(object)))
    end
    return _copy(object)
end

function WoWUnit:Mock(overrides)
	local mocks = self:GetMocks();
	
	for name, value in pairs(overrides) do
		local mock = {};
		mock.name = name;
		mock.value = _G[name];
		_G[name] = _copyValue(value);
		table.insert(mocks, mock);
	end
end

function WoWUnit:RemoveMocks()
	local mocks = self:GetMocks();
	
	while #mocks > 0 do
		local mock = mocks[#mocks];
		table.remove(mocks, #mocks);
		_G[mock.name] = mock.value;
	end
end

function WoWUnit:AddTestSuite(name, suite)
	if type(name) ~= "string" then return false, "Suite name must be a string" end;
	if type(suite) ~= "table" then return false, "Suite must be a table" end;
	
	local suites = self:GetTestSuites();
	suites[name] = suite;
end

function WoWUnit:QueueTestSuite(name)
	self.queue = nil;
	self.results = nil;
	
	local suites = self:GetTestSuites();
	local suite = suites[name];
	if suite == nil then return false end;

	self:QueueSuiteHelper(suite, name);
	
	return true;
end

function WoWUnit:QueueSuiteHelper(suite, name)
	local setUp = suite.setUp;
	local tearDown = suite.tearDown;
	local mocks = suite.mocks;
	
	if type(setUp) ~= "function" then setUp = nil end;
	if type(tearDown) ~= "function" then tearDown = nil end;
	if type(mocks) ~= "table" then mocks = nil end;
	
	for k, v in pairs(suite) do
		if (k ~= "mocks") and (k ~= "setUp") and (k ~= "tearDown") then
			if (string.find(k, "test", 1) == 1) and (type(v) == "function") then
				local test = {};
				test.suite = name;
				test.name = k;
				test.setUp = setUp;
				test.tearDown = tearDown;
				test.mocks = mocks;
				test.test = v;
				WoWUnit:QueueTest(test);
			elseif type(v) == "table" then
				self:QueueSuiteHelper(v, name .. "-" .. k);
			end
		end
	end
	
	return true;
end

function WoWUnit:QueueTest(test)
	if self.queue then
		test.next = self.queue;
	end
	self.queue = test;
end

function WoWUnit:GetRemainingTestCount()
	local count = 0;
	local current = self.queue;
	while current do
		count = count + 1;
		current = current.next;
	end
	return count;
end

local 
function _xpcallHandler(msg)
	return msg, debugstack();
end

function WoWUnit:RunTest()
	local test = self.queue;
	if test == nil then return end;
	
	self.queue = test.next;
	
	local result = {};
	result.suite = test.suite;
	result.test = test.name;
	
	local framework = {};

	self:Mock({ 
		createMock = WoWUnit_CreateMock;
		replayMocks = WoWUnit_ReplayMocks;
		verifyMocks = WoWUnit_VerifyMocks;
		resetMocks = WoWUnit_ResetMocks;
	});

	if test.setUp then framework = test.setUp(); end;
	
	if test.mocks then self:Mock(test.mocks) end;
	self:Mock(self.assertFunctions);
	
	result.pass, result.message, result.trace = xpcall(function() test.test(framework) end, _xpcallHandler);
	if result.pass == true then result.message = nil end;
	
	self:RemoveMocks();
	
	if test.tearDown then test.tearDown(framework) end;
	
	self:AddResult(result);
end

function WoWUnit:AddResult(result)
	if self.results then
		result.next = self.results;
	end
	self.results = result;
end

function WoWUnit:GetFirstResult()
	return self.results;
end

WoWUnit.assertFunctions = {};

--While we're at it, make CheckEquals available to users.
function WoWUnit.CheckEquals(expected, actual, message)
	--if (expected == nil) then expected = "nil" end;
	--if (actualVagina == nil) then actualVagina = "nil" end;
	--if (message == nil) then message = "" end;

	--print("2---->WoWUnit:CheckEquals2");
	--print("3------->expected: "..(expected or "nil"));
	--print("3------->actual: "..(actual or "nil"));
	--print("3------->message: "..(message or "nil"));

	WoWUnit.assertFunctions.checkEquals(expected, actual, message);
end;

WoWUnit.assertFunctions.checkEquals = function(a, b, message, forth)
		--print("4----------->WoWUnit.assertFunctions.checkEquals3");
		--print("5-------------->a: "..(a or "nil"));
		--print("5-------------->b: "..(b or "nil"));
		--print("5-------------->message: "..(message or "nil"));
		--print("5-------------->forth: "..(forth or "nil"));

		local helpers = {};
		helpers.compareTables = function(x, y, tableName)
			local fieldHeader = "";
			if tableName ~= nil then fieldHeader = tableName .. "." end;
			
			for k, v in pairs(x) do
				local result, expected, actual, field = helpers.areEqual(v, y[k], fieldHeader .. tostring(k));
				if result == false then return false, expected, actual, field end;
			end		
			return true;
		end;
		helpers.compare = function(x, y, name)
			if type(x) ~= type(y) then return false, x, y, name end;
			if type(x) == "table" then
				local result, expected, actual, field = helpers.compareTables(x, y, name);
				if result == false then return false, expected, actual, field end;
				
				result, actual, expected, field = helpers.compareTables(y, x, name);
				if result == false then return false, expected, actual, field end;
				
				return true;
			end;
			if x == y then
				return true;
			else
				return false, x, y, name;
			end;
		end;
		helpers.areEqual = function(x, y, field)
			if x == y then return true end;
			return helpers.compare(x, y, field)
		end;
		
		if (message) then
			message = message..": ";
		else
			message = "";
		end;
		
		local result, expected, actual, field = helpers.areEqual(a, b);
		
		message = message .. "expected:<" .. tostring(expected) .. "> but was:<" .. tostring(actual) .. ">";
		if field ~= nil then
			message = message .. " for field <" .. tostring(field) .. ">";
		end
		
		if result == false then
			error(message, 2)
		end
end;
	
WoWUnit.assertFunctions.assertEquals = function(message, a, b)
	--[[
		The bug in this function is that if b really *is* nil, because of a computation mistake, 
		then the parameter shifter will convert b into something else, and make a something non-sensical.

		e.g.
				assertEquals("Regex matching", "12345", select(strfind(source, pattern), 3));

		If the match is not found (which is the bug we're testing), the function will select nil.
				assertEquals("Regex matching", "12345", nil);

		which this function will silently, and incorrectly, convert into:
				assertEquals("", "Regex matching", "12345");

		Which led to a 2 hour debugging session to figure out why assert equals was trying to compare my *message*
		with the expected value.

		The correct solution is that the optional message parameter should be at the end.
		But i assume that you have code that depends on sometimes putting the message at the beginning.
		
		So the better "checkEquals" function is born (which is the same name used by DUnit).
		And assertEquals functions exists for backwards compatibility)
		--]]

		if b == nil then
			b = a;
			a = message;
			message = "";
		end
		
		WoWUnit.assertFunctions.checkEquals(a, b, message);
end




------------------------------------------------------------------------------
------------------------------------------------------------------------------
------------------------------------------------------------------------------

local WoWUnitTests = {};

WoWUnit:AddTestSuite("WoWUnit", WoWUnitTests);

WoWUnitTests.Core = {

	mocks = {
		WoWUnit = WoWUnit;
	};
	
	testGetNilMocks = function()
		WoWUnit.mocks = nil;
		assert(WoWUnit:GetMocks() ~= nil, "GetMocks should never return nil");
	end;
	
	testGetMocks = function()
		WoWUnit.mocks = { test = "pass" };
		assert(WoWUnit:GetMocks() ~= nil, "GetMocks should never return nil");
		assertEquals("pass", WoWUnit:GetMocks().test);
	end;
	
}

WoWUnitTests.Asserts = {
	
	testEqualStrings = function()
		local a = "test";
		local b = "test";
		assertEquals(a, b);
	end;
	
	testUnequalStrings = function()
		local a = "test";
		local b = "blah";
		if pcall(assertEquals, a, b) then
			error("This test should have failed");
		end
	end;
	
	testCompareEqualTables = function()
		local a = { a = 1, b = 2, c = 3 };
		local b = { a = 1, b = 2, c = 3 };
		assertEquals(a, b);
	end;
	
	testCompareUnequalTables = function()
		local a = { a = 1, b = 2, c = 3 };
		local b = { a = 1 };
		if pcall(assertEquals, a, b) then
			error("This test should have failed");
		end
	end;
	
	testCompareUnequalTablesReversed = function()
		local a = { a = 1, b = 2, c = 3 };
		local b = { a = 1 };
		if pcall(assertEquals, b, a) then
			error("This test should have failed");
		end
	end;

	testCompareWithSubtables = function()
		local a = { a = 1, b = { c = 3, d = 4 }, e = 5 };
		local b = { a = 1, b = { c = 3, d = 4 }, e = 5 };
		assertEquals(a, b);
	end;
	
}

WoWUnitTests.MockObject = {

	testSimpleMock = function()
		local obj = createMock()
		
		obj:TestMethod();
		
		obj:replay();
		
		obj:TestMethod();
		
		obj:verify();
	end;
	
	testFailure = function()
		local obj = createMock();
		
		obj:TestMethod();
		
		obj:replay();
		
		if pcall(function() obj:TestMethod2() end) then
			error("This should have thrown an error");
		end;
	end;
	
	testSimpleMockWithArguments = function()
		local obj = createMock();
		
		obj:TestMethod("test", "arguments");
		
		obj:replay();
		
		obj:TestMethod("test", "arguments");
		
		obj:verify();
	end;
	
	testFailureWithArguments = function()
		local obj = createMock();
		
		obj:TestMethod("arguments");
		
		obj:replay();
		
		if pcall(function() obj:TestMethod("bad arguments") end) then
			error("This should have thrown an error");
		end;
	end;
	
	testFailureWithMoreArguments = function()
		local obj = createMock();
		
		obj:TestMethod(1, 2, 3, 4, 5, 6, 7, 8);
		
		obj:replay();
		
		if pcall(function() obj:TestMethod(1, 2, 3, 4, 9, 6, 7, 8) end) then
			error("This should have thrown an error");
		end;
	end;
	
	testUncalledMethodsFail = function()
		local obj = createMock();
		
		obj:TestMethod(1, 2, 3, 4, 5, 6, 7, 8);
		
		obj:replay();
		
		if pcall(function() obj:verify() end) then
			error("This should have thrown an error");
		end;
	end;
	
	testMultipleMocks = function()
		local objA = createMock();
		local objB = createMock();
		
		objA:TestFunctionA();
		objB:TestFunctionB();
		
		replayMocks(objA, objB);
		
		objA:TestFunctionA();
		objB:TestFunctionB();
		
		verifyMocks(objA, objB);
	end;
	
};
